from django.utils.translation import gettext_lazy as _

from trench.backends import AbstractMessageDispatcher


class YubiKeyBackend(AbstractMessageDispatcher):
    SMS_BODY = _('Your verification code is: ')

    def dispatch_message(self):
        """
        Do nothing.
        """

        return {'message': _('Generate code using YubiKey')}  # pragma: no cover # noqa



from django.utils.six import text_type

from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework_simplejwt.tokens import RefreshToken

from trench.views.base import MFACodeLoginMixin, MFACredentialsLoginMixin


class ObtainJSONWebTokenMixin:
    def handle_user_login(self, request, serializer, *args, **kwargs):
        token = RefreshToken.for_user(serializer.user)
        return Response(
            {
                'refresh': text_type(token),
                'access': text_type(token.access_token)
            }
        )


class JSONWebTokenLoginOrRequestMFACode(MFACredentialsLoginMixin,
                                        ObtainJSONWebTokenMixin,
                                        GenericAPIView):
    pass


class JSONWebTokenLoginWithMFACode(MFACodeLoginMixin,
                                   ObtainJSONWebTokenMixin,
                                   GenericAPIView):
    pass



from trench.backends import AbstractMessageDispatcher
from trench.utils import create_qr_link


class ApplicationBackend(AbstractMessageDispatcher):
    def dispatch_message(self):
        return {'qr_link': create_qr_link(self.obj.secret, self.user)}



from django.contrib.auth import authenticate, get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db import models as django_models
from django.db.utils import DatabaseError
from django.utils.translation import ugettext as _

from rest_framework import fields, serializers

from trench.settings import api_settings
from trench.utils import (
    create_secret,
    get_mfa_model,
    get_nested_attr,
    set_nested_attr,
    user_token_generator,
    validate_backup_code,
    validate_code,
)


User = get_user_model()
MFAMethod = get_mfa_model()

mfa_methods_items = api_settings.MFA_METHODS.items()
MFA_METHODS = [
    (k, v.get('VERBOSE_NAME', _(k))) for k, v in mfa_methods_items
]


class RequestMFAMethodActivationSerializer(serializers.Serializer):
    serializer_field_mapping = {
        django_models.AutoField: fields.IntegerField,
        django_models.BigIntegerField: fields.IntegerField,
        django_models.BooleanField: fields.BooleanField,
        django_models.CharField: fields.CharField,
        django_models.CommaSeparatedIntegerField: fields.CharField,
        django_models.DateField: fields.DateField,
        django_models.DateTimeField: fields.DateTimeField,
        django_models.DecimalField: fields.DecimalField,
        django_models.EmailField: fields.EmailField,
        django_models.Field: fields.ModelField,
        django_models.FileField: fields.FileField,
        django_models.FloatField: fields.FloatField,
        django_models.ImageField: fields.ImageField,
        django_models.IntegerField: fields.IntegerField,
        django_models.NullBooleanField: fields.NullBooleanField,
        django_models.PositiveIntegerField: fields.IntegerField,
        django_models.PositiveSmallIntegerField: fields.IntegerField,
        django_models.SlugField: fields.SlugField,
        django_models.SmallIntegerField: fields.IntegerField,
        django_models.TextField: fields.CharField,
        django_models.TimeField: fields.TimeField,
        django_models.URLField: fields.URLField,
        django_models.GenericIPAddressField: fields.IPAddressField,
        django_models.FilePathField: fields.FilePathField,
    }

    default_error_messages = {
        'required_field_missing':
        _('Required field not provided'),
        'required_field_update_failed':
        _('Failed to update required User data. Try again.')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        context = kwargs['context']

        self.user = context['request'].user
        self.conf = api_settings.MFA_METHODS[context['name']]

        self.source_field = self.conf.get('SOURCE_FIELD')
        if self.source_field:
            value, field_name, klass = get_nested_attr(
                self.user, self.source_field)
            if not value:
                self.fields[field_name] = (
                    self.get_serializer_field_mapping()[klass](
                        required=True)
                )
                self.required_field_name = field_name

    def validate(self, attrs):
        if hasattr(self, 'required_field_name'):
            if self.required_field_name in attrs:
                try:
                    set_nested_attr(
                        self.user,
                        self.source_field,
                        attrs[self.required_field_name],
                    )
                except (AttributeError, DatabaseError):  # pragma: no cover
                    self.fail('required_field_update_failed')  # pragma: no cover  # noqa
            else:  # pragma: no cover
                self.fail('required_field_missing')  # pragma: no cover
        return attrs

    def create(self, validated_data):
        """
        Creates new MFAMethod object for given user, sets it as inactive,
        and marks as primary if no other active MFAMethod exists for user.
        """
        return MFAMethod.objects.get_or_create(
            user=self.user,
            name=self.context['name'],
            defaults={
                'secret': create_secret(),
                'is_active': False,
            }
        )

    def get_serializer_field_mapping(self):
        return self.serializer_field_mapping


class ProtectedActionSerializer(serializers.Serializer):
    requires_mfa_code = None
    code = serializers.CharField(
        required=False,
    )

    default_error_messages = {
        'otp_code_missing': _('OTP code not provided.'),
        'code_invalid_or_expired': _('Code invalid or expired.'),
    }

    def _validate_code(self, value):
        if not value:
            self.fail('otp_code_missing')

        obj = self.context['obj']
        validity_period = (
            self.context['conf'].get('VALIDITY_PERIOD')
            or api_settings.DEFAULT_VALIDITY_PERIOD  # noqa
        )
        validated_backup_code = validate_backup_code(value, obj.backup_codes)
        if validate_code(value, obj, validity_period):
            return value
        if validated_backup_code:
            obj.remove_backup_code(validated_backup_code)
            return value

        self.fail('code_invalid_or_expired')

    def validate(self, data):
        if self.requires_mfa_code:
            self._validate_code(data.get('code'))

        return super().validate(data)


class RequestMFAMethodActivationConfirmSerializer(ProtectedActionSerializer):
    requires_mfa_code = True


class RequestMFAMethodDeactivationSerializer(ProtectedActionSerializer):
    requires_mfa_code = api_settings.CONFIRM_DISABLE_WITH_CODE

    default_error_messages = {
        'code_invalid_or_expired': _('Code invalid or expired.'),
        'new_primary_same_as_old':
        _(
            'MFA Method to be deactivated cannot be chosen'
            ' as new primary method.'
        ),
        'method_not_registered_for_user':
        _(
            'Selected new primary MFA method'
            ' is not registered for current user.'
        ),
        'new_primary_method_inactive':
        _('MFA Method selected as new primary method is not active'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        context = kwargs['context']
        self.user = context['request'].user

        is_current_method_primary = (
            MFAMethod.objects
            .filter(user=self.user, name=context['name'])
            .get()
            .is_primary
        )

        self.users_active_methods_count = (
            MFAMethod.objects
            .filter(user=self.user, is_active=True)
            .count()
        )
        if is_current_method_primary and self.users_active_methods_count > 2:
            self.fields['new_primary_method'] = serializers.CharField(
                max_length=255, required=True)
        else:
            self.new_method = None

    def validate_new_primary_method(self, value):
        method_to_deactivate = self.context.get('name')

        if method_to_deactivate == value:
            self.fail('new_primary_same_as_old')

        try:
            self.new_method = MFAMethod.objects.get(user=self.user, name=value)
        except MFAMethod.DoesNotExist:
            self.fail('method_not_registered_for_user')
        if not self.new_method.is_active:
            self.fail('new_primary_method_inactive')  # pragma: no cover
        return value


class RequestMFAMethodBackupCodesRegenerationSerializer(
    ProtectedActionSerializer
):
    requires_mfa_code = api_settings.CONFIRM_BACKUP_CODES_REGENERATION_WITH_CODE  # noqa


class RequestMFAMethodCodeSerializer(serializers.Serializer):
    method = serializers.CharField(
        max_length=255,
        required=False,
    )

    default_error_messages = {
        'mfa_method_not_exists': _('Requested MFA method does not exists'),
    }

    def validate_method(self, value):
        if value and value not in api_settings.MFA_METHODS:
            self.fail('mfa_method_not_exists')
        return value


class LoginSerializer(serializers.Serializer):
    """
    Validates user's credentials.
    """
    password = serializers.CharField(
        style={'input_type': 'password'},
        write_only=True,
    )

    def __init__(self, *args, **kwargs):
        super(LoginSerializer, self).__init__(*args, **kwargs)
        self.user = None
        self.fields[User.USERNAME_FIELD] = serializers.CharField()

    def validate(self, attrs):
        self.user = authenticate(
            request=self.context.get('request'),
            username=attrs.get(User.USERNAME_FIELD),
            password=attrs.get('password')
        )

        if not getattr(self.user, api_settings.USER_ACTIVE_FIELD, True):
            msg = _('User account is disabled.')  # pragma: no cover
            raise serializers.ValidationError(msg)  # pragma: no cover

        if not self.user:
            msg = _('Unable to login with provided credentials.')
            raise serializers.ValidationError(msg)

        return {}


class CodeLoginSerializer(serializers.Serializer):
    """
    Validates given token and OTP code.
    """
    ephemeral_token = serializers.CharField()
    code = serializers.CharField()

    default_error_messages = {
        'invalid_token': _('Invalid or expired token.'),
        'invalid_code': _('Invalid or expired code.'),
    }

    def validate(self, attrs):
        ephemeral_token = attrs.get('ephemeral_token')
        code = attrs.get('code')

        self.user = user_token_generator.check_token(ephemeral_token)
        if not self.user:
            self.fail('invalid_token')

        for auth_method in self.user.mfa_methods.filter(is_active=True):
            validated_backup_code = validate_backup_code(
                code,
                auth_method.backup_codes,
            )
            if validate_code(code, auth_method):
                return attrs
            if validated_backup_code:
                auth_method.remove_backup_code(validated_backup_code)
                return attrs

        self.fail('invalid_code')


class UserMFAMethodSerializer(serializers.ModelSerializer):
    """
    Serializes active MFA method for user preview
    """
    class Meta:
        model = MFAMethod
        fields = ('name', 'is_primary')


class ChangePrimaryMethodSerializer(serializers.Serializer):
    """
    Serializes request to change default authentication method.
    """
    code = serializers.CharField()
    method = serializers.ChoiceField(choices=MFA_METHODS)

    default_error_messages = {
        'not_enabled': _('2FA is not enabled.'),
        'invalid_code': _('Invalid or expired code.'),
        'missing_method': _('Target method does not exist or is not active'),
    }

    def validate(self, attrs):
        user = self.context.get('request').user
        try:
            current_method = user.mfa_methods.get(
                is_primary=True,
                is_active=True,
            )
        except ObjectDoesNotExist:
            self.fail('not_enabled')
        try:
            new_primary_method = user.mfa_methods.get(
                name=attrs.get('method'),
                is_active=True,
            )
        except ObjectDoesNotExist:
            self.fail('missing_method')
        code = attrs.get('code')
        validated_backup_code = validate_backup_code(
            code,
            current_method.backup_codes,
        )
        if validate_code(code, current_method):
            attrs.update(new_method=new_primary_method)
            attrs.update(old_method=current_method)
            return attrs
        elif validated_backup_code:
            attrs.update(new_method=new_primary_method)
            attrs.update(old_method=current_method)
            current_method.remove_backup_code(validated_backup_code)
            return attrs
        else:
            self.fail('invalid_code')

    def save(self):
        new_method = self.validated_data.get('new_method')
        old_method = self.validated_data.get('old_method')
        new_method.is_primary = True
        old_method.is_primary = False
        new_method.save()
        old_method.save()



class BaseMFAException(Exception):
    pass


class MFADoesNotExist(BaseMFAException):
    """User does not have MFA set up"""
    pass


class MFAAlreadyExist(BaseMFAException):
    """User already have MFA set up"""
    pass


class InvalidOTPCode(BaseMFAException):
    """Provided OTP code is not valid"""
    pass


class MethodNotAllowed(BaseMFAException):
    """Given MFA method is not allowed"""
    pass


class MissingSourceFieldAttribute(BaseMFAException):
    """Atrribute specified as required for given backend is missing"""
    pass



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _

from rest_framework.settings import APISettings, perform_import


class TrenchAPISettings(APISettings):

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'TRENCH_AUTH', {})
        return self._user_settings

    def __getattr__(self, attr):

        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)
        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Validations for attrs
        if attr == 'BACKUP_CODES_CHARACTERS' and ',' in val:
            raise ImproperlyConfigured(
                'Cannot use , as a character for {}.'.format(attr),
            )
        if attr == 'MFA_METHODS':
            for method_name, method_config in val.items():
                if 'HANDLER' not in method_config:
                    raise ImproperlyConfigured(
                        'Missing handler in {} configuration.'.format(
                            method_name
                        ),
                    )
                method_config['HANDLER'] = perform_import(
                    method_config.get('HANDLER'),
                    'HANDLER',
                )
        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def __getitem__(self, attr):
        return self.__getattr__(attr)


DEFAULTS = {
    'FROM_EMAIL': getattr(settings, 'DEFAULT_FROM_EMAIL'),
    'USER_MFA_MODEL': 'trench.MFAMethod',
    'USER_ACTIVE_FIELD': 'is_active',
    'BACKUP_CODES_QUANTITY': 5,
    'BACKUP_CODES_LENGTH': 6,  # keep (quantity * length) under 200
    'BACKUP_CODES_CHARACTERS': (
        '0123456789'
    ),
    'DEFAULT_VALIDITY_PERIOD': 30,
    'CONFIRM_DISABLE_WITH_CODE': False,
    'CONFIRM_BACKUP_CODES_REGENERATION_WITH_CODE': True,
    'ALLOW_BACKUP_CODES_REGENERATION': True,
    'ENCRYPT_BACKUP_CODES': True,
    'APPLICATION_ISSUER_NAME': 'MyApplication',
    'CACHE_PREFIX': 'trench',
    'MAX_METHODS_PER_USER': 3,
    'MFA_METHODS': {
        'sms_twilio': {
            'VERBOSE_NAME': _('sms_twilio'),
            'VALIDITY_PERIOD': 60 * 10,
            'HANDLER': 'trench.backends.twilio.TwilioBackend',
            'SOURCE_FIELD': 'phone_number',
            'TWILIO_ACCOUNT_SID': 'YOUR KEY',
            'TWILIO_AUTH_TOKEN': 'YOUR KEY',
            'TWILIO_VERIFIED_FROM_NUMBER': 'YOUR TWILIO REGISTERED NUMBER',
        },
        'sms_api': {
            'VERBOSE_NAME': _('sms_api'),
            'VALIDITY_PERIOD': 60 * 10,
            'HANDLER': 'trench.backends.sms_api.SmsAPIBackend',
            'SOURCE_FIELD': 'phone_number',
            'SMSAPI_ACCESS_TOKEN': 'YOUR SMSAPI TOKEN',
            'SMSAPI_FROM_NUMBER': 'YOUR REGISTERED NUMBER',
        },
        'email': {
            'VERBOSE_NAME': _('email'),
            'VALIDITY_PERIOD': 60 * 10,
            'HANDLER': 'trench.backends.basic_mail.SendMailBackend',
            'SOURCE_FIELD': 'email',
        },
        'app': {
            'VERBOSE_NAME': _('app'),
            'VALIDITY_PERIOD': 60 * 10,
            'USES_THIRD_PARTY_CLIENT': True,
            'HANDLER': 'trench.backends.application.ApplicationBackend',
        },
        'yubi': {
            'VERBOSE_NAME': _('yubi'),
            'VALIDITY_PERIOD': 60 * 10,
            'SOURCE_FIELD': 'yubikey_id',
            'HANDLER': 'trench.backends.yubikey.YubiKeyBackend',
            'YUBICLOUD_CLIENT_ID': 'YOUR KEY',
        },
    },
}

api_settings = TrenchAPISettings(None, DEFAULTS, None)



from datetime import datetime

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import check_password
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.core.exceptions import FieldDoesNotExist
from django.utils.crypto import (
    constant_time_compare,
    get_random_string,
    salted_hmac,
)
from django.utils.http import base36_to_int, int_to_base36

import pyotp
from yubico_client.yubico import Yubico
from yubico_client.yubico_exceptions import YubicoError

from trench.settings import api_settings


User = get_user_model()


class UserTokenGenerator(PasswordResetTokenGenerator):
    """
    Custom token generator:
        - user pk in token
        - expires after 15 minutes
        - longer hash (40 instead of 20)
    """
    key_salt = 'django.contrib.auth.tokens.PasswordResetTokenGenerator'
    secret = settings.SECRET_KEY

    def make_token(self, user):
        return self._make_token_with_timestamp(
            user, int(datetime.now().timestamp())
        )

    def check_token(self, token):
        if not token:
            return None

        try:
            token = str(token)
            user_pk, ts_b36, hash = token.split('-')
            ts = base36_to_int(ts_b36)
            user = User._default_manager.get(pk=user_pk)
        except (ValueError, TypeError, User.DoesNotExist):
            return None

        if (datetime.now().timestamp() - ts) > (60 * 15):
            return None  # pragma: no cover

        if not constant_time_compare(
                self._make_token_with_timestamp(user, ts), token):
            return None

        return user

    def _make_token_with_timestamp(self, user, timestamp):
        ts_b36 = int_to_base36(timestamp)
        hash = salted_hmac(
            self.key_salt,
            self._make_hash_value(user, timestamp),
            secret=self.secret,
        ).hexdigest()
        return '%s-%s-%s' % (user.pk, ts_b36, hash)


user_token_generator = UserTokenGenerator()


def create_secret():
    """
    Creates new random secret.

    :returns: Random secret
    :rtype: str
    """

    return pyotp.random_base32(length=20)


def create_otp_code(secret):
    """
    Creates new OTP code.

    :param secret: Secret used to generate OTP
    :type secret: str

    :returns: OTP code
    :rtype: str
    """

    totp = pyotp.TOTP(secret)
    return totp.now()


def create_qr_link(secret, user):
    """
    Creates QR link to set application OTP.

    :param secret: Secret used to generate OTP
    :type secret: str
    :param user: User that link should be created for
    :type user: User

    :returns: Link to generate QR code from
    :rtype: str
    """

    totp = pyotp.TOTP(secret)
    return totp.provisioning_uri(
        getattr(user, User.USERNAME_FIELD),
        api_settings.APPLICATION_ISSUER_NAME,
    )


def generate_backup_codes(
    quantity=api_settings.BACKUP_CODES_QUANTITY,
    length=api_settings.BACKUP_CODES_LENGTH,
    allowed_chars=api_settings.BACKUP_CODES_CHARACTERS,
):
    """
    Generates random encrypted backup codes.

    :param quantity: How many codes should be generated
    :type quantity: int
    :param length: How long codes should be
    :type length: int
    :param allowed_chars: Characters to create backup codes from
    :type allowed_chars: str

    :returns: Encrypted backup codes
    :rtype: list
    """

    return [get_random_string(length, allowed_chars) for _ in range(quantity)]


def validate_code(
    code,
    mfa_method,
    valid_window=api_settings.DEFAULT_VALIDITY_PERIOD,
):

    if mfa_method.name == 'yubi':
        conf = api_settings.MFA_METHODS['yubi']
        client = Yubico(conf['YUBICLOUD_CLIENT_ID'])

        try:
            return client.verify(code, timestamp=True)

        except (YubicoError, Exception):
            return False

    return (
        pyotp
        .TOTP(mfa_method.secret)
        .verify(
            code,
            valid_window=int(valid_window / 30)
        )
    )


def get_mfa_model():
    return apps.get_model(api_settings.USER_MFA_MODEL)


def parse_dotted_path(path):
    """
    Extracts attribute name from dotted path.
    """
    try:
        objects, attr = path.rsplit('.', 1)
    except ValueError:
        objects = None
        attr = path

    return objects, attr


def get_innermost_object(obj, dotted_path=None):
    """
    For given object return innermost object.
    """
    if not dotted_path:
        return obj
    for o in dotted_path.split('.'):
        obj = getattr(obj, o)
    return obj  # pragma: no cover


def get_nested_attr(obj, path):
    """
    For attribute in dotted path notation retrieves its
    value, name, and field class.
    """
    objects, attr = parse_dotted_path(path)
    try:
        _obj = get_innermost_object(obj, objects)
    except AttributeError:  # pragma: no cover
        return (None, None, None)  # pragma: no cover

    try:
        field = _obj._meta.get_field(attr)
    except FieldDoesNotExist:  # pragma: no cover
        return (None, None, None)  # pragma: no cover

    return (
        field.value_from_object(_obj),
        field.get_attname(),
        field.__class__,
    )


def set_nested_attr(obj, path, value):
    """
    Sets value on attribute specified by path.
    Raises AttributeError, DatabaseError
    """
    objects, attr = parse_dotted_path(path)
    _obj = get_innermost_object(obj, objects)
    setattr(_obj, attr, value)
    _obj.save(update_fields=[attr])


def get_nested_attr_value(obj, path):
    objects, attr = parse_dotted_path(path)

    try:
        _obj = get_innermost_object(obj, objects)
    except AttributeError:  # pragma: no cover
        return None  # pragma: no cover

    return getattr(_obj, attr)


def validate_backup_code(value, backup_codes):
    """
    Validates provided code against list of hashed backup codes.
    Returns correct code.

    :param value:
    :param backup_codes:
    :return:
    """
    if api_settings.ENCRYPT_BACKUP_CODES:
        validated_codes = [_ for _ in backup_codes if check_password(value, _)]
    else:  # pragma: no cover
        validated_codes = [_ for _ in backup_codes if value in backup_codes]

    if validated_codes:
        return validated_codes[0]



from trench.utils import (
    get_nested_attr_value,
    create_otp_code,
)
from trench.exceptions import MissingSourceFieldAttribute


class AbstractMessageDispatcher:
    def __init__(self, user, obj, conf):
        self.user = user
        self.obj = obj
        self.conf = conf
        self.to = ''

        if 'SOURCE_FIELD' in conf:
            value = get_nested_attr_value(user, conf['SOURCE_FIELD'])
            if not value:
                raise MissingSourceFieldAttribute(  # pragma: no cover
                    'Could not retrieve attribute '
                    '{} for given user'.format(conf['SOURCE_FIELD'])
                )
            self.to = value

    def dispatch_message(self):
        pass  # pragma: no cover

    def create_code(self):
        return create_otp_code(self.obj.secret)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2019-01-11 14:03
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('trench', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='mfamethod',
            old_name='backup_codes',
            new_name='_backup_codes',
        ),
        migrations.AlterField(
            model_name='mfamethod',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mfa_methods', to=settings.AUTH_USER_MODEL, verbose_name='user'),
        ),
    ]



from django.contrib.auth.hashers import make_password
from django.db import IntegrityError, transaction
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _

from rest_framework import status
from rest_framework.generics import CreateAPIView, GenericAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from trench import serializers
from trench.settings import api_settings
from trench.utils import (
    generate_backup_codes,
    get_mfa_model,
    user_token_generator,
)


MFAMethod = get_mfa_model()
requires_encryption = api_settings.ENCRYPT_BACKUP_CODES


class MFACredentialsLoginMixin:
    """
    Mixin handling user log in. Checks if primary MFA method
    is active and dispatches code if so. Else calls handle_user_login.
    """
    serializer_class = serializers.LoginSerializer

    def handle_mfa_response(self, user, mfa_method, *args, **kwargs):
        data = {
            'ephemeral_token': user_token_generator.make_token(user),
            'method': mfa_method.name,
            'other_methods': serializers.UserMFAMethodSerializer(
                user.mfa_methods.filter(is_active=True, is_primary=False),
                many=True,
            ).data,
        }
        return Response(data)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.user
        auth_method = (
            user.mfa_methods
                .filter(is_primary=True, is_active=True)
                .first()
        )
        if auth_method:
            conf = api_settings.MFA_METHODS[auth_method.name]
            handler = conf['HANDLER'](
                user=user,
                obj=auth_method,
                conf=conf,
            )
            handler.dispatch_message()
            return self.handle_mfa_response(user, auth_method)

        return self.handle_user_login(
            request=request,
            serializer=serializer,
            *args,
            **kwargs
        )


class MFACodeLoginMixin:
    """
    Mixin handling user login if MFA auth is enabled.
    Expects ephemeral token and valid MFA code.
    Checks against all active MFA methods.
    """
    serializer_class = serializers.CodeLoginSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        return self.handle_user_login(
            request=request,
            serializer=serializer,
            *args,
            **kwargs
        )


class RequestMFAMethodActivationView(GenericAPIView):
    """
    View handling new MFA method activation requests.
    If validation passes, new MFAMethod (inactive) object
    is created.
    """
    serializer_class = serializers.RequestMFAMethodActivationSerializer
    permission_classes = (IsAuthenticated, )
    http_method_names = ['post']

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['name'] = self.kwargs.get('method')
        return context

    def post(self, request, *args, **kwargs):
        self.mfa_method_name = kwargs.get('method')

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj, created = serializer.save()
        if not created and obj.is_active:
            return Response(
                {'error': 'MFA method already active.'},
                status=status.HTTP_400_BAD_REQUEST,
            )

        conf = api_settings.MFA_METHODS[self.mfa_method_name]

        handler = conf['HANDLER'](
            user=request.user,
            obj=obj,
            conf=conf,
        )
        return Response(handler.dispatch_message(), status=status.HTTP_200_OK)


class RequestMFAMethodActivationConfirmView(GenericAPIView):
    serializer_class = serializers.RequestMFAMethodActivationConfirmSerializer
    permission_classes = (IsAuthenticated,)
    http_method_names = ['post']

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update({
            'name': self.kwargs['method'],
            'obj': self.obj,
            'conf': api_settings.MFA_METHODS[self.kwargs['method']],
        })
        return context

    def post(self, request, *args, **kwargs):
        self.mfa_method_name = self.kwargs['method']
        self.obj = get_object_or_404(
            MFAMethod, user=request.user, name=self.mfa_method_name)

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        backup_codes = generate_backup_codes()

        if requires_encryption:
            self.obj.backup_codes = [
                make_password(backup_code) for backup_code in backup_codes
            ]
        else:  # pragma: no cover
            self.obj.backup_codes = backup_codes

        self.obj.is_active = True
        self.obj.is_primary = not MFAMethod.objects.filter(
            user=request.user,
            is_active=True,
        ).exists()
        self.obj.save(
            update_fields=['is_active', '_backup_codes', 'is_primary']
        )

        return Response({'backup_codes': backup_codes})


class RequestMFAMethodDeactivationView(GenericAPIView):
    serializer_class = serializers.RequestMFAMethodDeactivationSerializer
    permission_classes = (IsAuthenticated,)
    http_method_names = ['post']

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update({
            'name': self.kwargs['method'],
            'obj': self.obj,
            'conf': api_settings.MFA_METHODS[self.kwargs['method']],
        })
        return context

    def post(self, request, *args, **kwargs):
        self.mfa_method_name = kwargs.get('method')
        self.obj = get_object_or_404(
            MFAMethod, user=request.user, name=self.mfa_method_name)

        if not self.obj.is_active:
            return Response(
                {'error': _('Method already disabled.')},
                status=status.HTTP_400_BAD_REQUEST,
            )

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        try:
            with transaction.atomic():
                default_update_fields = ['is_active']

                if serializer.users_active_methods_count >= 2:
                    new_primary_obj = (
                        getattr(serializer, 'new_method')
                        or MFAMethod.objects  # noqa
                        .filter(user=request.user, is_active=True)
                        .exclude(id=self.obj.id)
                        .first()
                    )

                    new_primary_obj.is_primary = True
                    new_primary_obj.save(update_fields=['is_primary'])

                default_update_fields.append('is_primary')
                self.obj.is_primary = False
                self.obj.is_active = False
                self.obj.save(update_fields=default_update_fields)
        except IntegrityError:  # pragma: no cover
            return Response(  # pragma: no cover
                {
                    'error': _('Failed to update MFA information')
                },  # pragma: no cover
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )
        return Response(status=status.HTTP_204_NO_CONTENT)


class RequestMFAMethodBackupCodesRegenerationView(GenericAPIView):
    serializer_class = serializers.RequestMFAMethodBackupCodesRegenerationSerializer  # noqa
    permission_classes = (IsAuthenticated,)
    http_method_names = ['post']

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update({
            'name': self.kwargs['method'],
            'obj': self.obj,
            'conf': api_settings.MFA_METHODS[self.kwargs['method']],
        })
        return context

    def post(self, request, *args, **kwargs):
        self.mfa_method_name = kwargs.get('method')
        self.obj = get_object_or_404(
            MFAMethod, user=request.user, name=self.mfa_method_name)

        if not self.obj.is_active:
            return Response(
                {'error': _('Method is disabled.')},
                status=status.HTTP_400_BAD_REQUEST,
            )

        serializer = self.get_serializer(data=request.data,)
        serializer.is_valid(raise_exception=True)

        backup_codes = generate_backup_codes()

        if requires_encryption:
            self.obj.backup_codes = [
                make_password(backup_code) for backup_code in backup_codes
            ]
        else:  # pragma: no cover
            self.obj.backup_codes = backup_codes

        self.obj.save(update_fields=['_backup_codes'])
        return Response({'backup_codes': backup_codes})


class GetMFAConfig(APIView):
    def get(self, request, *args, **kwargs):
        available_methods = [(k, v.get('VERBOSE_NAME'))
                             for k, v in api_settings.MFA_METHODS.items()]

        return Response(
            {
                'methods': available_methods,
                'confirm_disable_with_code': api_settings.CONFIRM_DISABLE_WITH_CODE,  # noqa
                'confirm_regeneration_with_code': api_settings.CONFIRM_BACKUP_CODES_REGENERATION_WITH_CODE,  # noqa
                'allow_backup_codes_regeneration': api_settings.ALLOW_BACKUP_CODES_REGENERATION,  # noqa
            },
            status=status.HTTP_200_OK,
        )


class ListUserActiveMFAMethods(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        active_mfa_methods = MFAMethod.objects.filter(
            user=request.user, is_active=True)
        serializer = serializers.UserMFAMethodSerializer(
            active_mfa_methods, many=True)
        return Response(serializer.data)


class RequestMFAMethodCode(GenericAPIView):
    serializer_class = serializers.RequestMFAMethodCodeSerializer
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        mfa_method_name = serializer.validated_data.get('method')
        if mfa_method_name:
            obj = get_object_or_404(
                MFAMethod,
                user=request.user,
                name=mfa_method_name,
                is_active=True,
            )

        conf = api_settings.MFA_METHODS.get(mfa_method_name)

        if not conf:
            return Response(  # pragma: no cover
                {'error', _('Requested MFA method does not exists')},
                status=status.HTTP_400_BAD_REQUEST,
            )

        handler = conf.get('HANDLER')(
            user=request.user,
            obj=obj,
            conf=conf,
        )
        dispatcher_resp = handler.dispatch_message()
        return Response(dispatcher_resp)


class ChangePrimaryMethod(CreateAPIView):
    serializer_class = serializers.ChangePrimaryMethodSerializer

    def post(self, request, *args, **kwargs):
        super().post(request, *args, **kwargs)

        return Response(
            serializers.UserMFAMethodSerializer(
                request.user.mfa_methods.filter(is_active=True),
                many=True,
            ).data,
            status=status.HTTP_200_OK,
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-10-12 11:34
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MFAMethod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='name')),
                ('secret', models.CharField(max_length=20, verbose_name='secret')),
                ('is_primary', models.BooleanField(default=False, verbose_name='is primary')),
                ('is_active', models.BooleanField(default=False, verbose_name='is active')),
                ('backup_codes', models.CharField(blank=True, max_length=255, verbose_name='backup codes')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mfa', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'MFA Method',
                'verbose_name_plural': 'MFA Methods',
            },
        ),
    ]



from django.conf import settings
from django.utils.translation import gettext_lazy as _

from twilio.base.exceptions import TwilioRestException
from twilio.rest import Client

from trench.backends import AbstractMessageDispatcher


class TwilioBackend(AbstractMessageDispatcher):
    SMS_BODY = _('Your verification code is: ')

    def dispatch_message(self):
        """
        Sends a SMS with verification code.
        :return:
        """
        code = self.create_code()

        if settings.DEBUG:
            try:  # pragma: no cover
                self.send_sms(self.to, code)  # pragma: no cover
            except TwilioRestException as e:  # pragma: no cover
                print(  # pragma: no cover
                    'Error found: {}\n'
                    'SMS to number {}\n'
                    'Your code is {}.'.format(e, self.to, code))
        else:
            self.send_sms(self.to, code)
        return {'message': _('SMS message with MFA code has been sent.')}  # pragma: no cover # noqa

    def send_sms(self, user_mobile, code):
        client = self.provider_auth()
        client.messages.create(  # pragma: no cover
            body=self.SMS_BODY + code,
            to=user_mobile,
            from_=self.conf.get('TWILIO_VERIFIED_FROM_NUMBER')
        )

    def provider_auth(self):
        return Client(
            self.conf.get('TWILIO_ACCOUNT_SID'),
            self.conf.get('TWILIO_AUTH_TOKEN'),
        )



from djoser.views import TokenCreateView, TokenDestroyView
from rest_framework.generics import GenericAPIView

from trench.views.base import MFACodeLoginMixin, MFACredentialsLoginMixin


class ObtainAuthTokenMixin:
    def handle_user_login(self, serializer, *args, **kwargs):
        return self._action(serializer)


class AuthTokenLoginOrRequestMFACode(MFACredentialsLoginMixin,
                                     ObtainAuthTokenMixin,
                                     TokenCreateView,
                                     GenericAPIView):
    pass


class AuthTokenLoginWithMFACode(MFACodeLoginMixin,
                                ObtainAuthTokenMixin,
                                GenericAPIView):
    pass


AuthTokenLogoutView = TokenDestroyView



from django.utils.translation import gettext_lazy as _

from smsapi.client import SmsApiPlClient

from trench.backends import AbstractMessageDispatcher


class SmsAPIBackend(AbstractMessageDispatcher):
    SMS_BODY = _('Your verification code is: ')

    def dispatch_message(self):
        """
        Sends a SMS with verification code.
        """

        code = self.create_code()
        self.send_sms(self.to, code)

        return {
            'message': _('SMS message with MFA code has been sent.')
        }  # pragma: no cover # noqa

    def send_sms(self, user_mobile, code):
        client = self.provider_auth()

        kwargs = {}
        if self.conf.get('SMSAPI_FROM_NUMBER'):
            kwargs['from_'] = self.conf.get(
                'SMSAPI_FROM_NUMBER'
            )  # pragma: no cover

        client.sms.send(
            message=self.SMS_BODY + code,
            to=user_mobile,
            **kwargs
        )

    def provider_auth(self):
        return SmsApiPlClient(
            access_token=self.conf.get('SMSAPI_ACCESS_TOKEN')
        )



from django.conf.urls import url

from trench.views import simplejwt as views


urlpatterns = [
    url(
        r'^login/$',
        views.JSONWebTokenLoginOrRequestMFACode.as_view(),
        name='generate-code',
    ),
    url(
        r'^login/code/',
        views.JSONWebTokenLoginWithMFACode.as_view(),
        name='generate-token',
    ),
]



from django.conf import settings
from django.conf.urls import url
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext_lazy as _

from trench.views import authtoken as views


__all__ = [
    'urlpatterns',
]

if 'djoser' not in settings.INSTALLED_APPS:
    msg = _(
        'Djoser not found in INSTALLED_APPS. '
        'Make sure you\'ve installed it, and '
        'add appropriate entry in settings.'
    )
    raise ImproperlyConfigured(msg)

urlpatterns = [
    url(
        r'^login/$',
        views.AuthTokenLoginOrRequestMFACode.as_view(),
        name='mfa-authtoken-login',
    ),
    url(
        r'^login/code/',
        views.AuthTokenLoginWithMFACode.as_view(),
        name='mfa-authtoken-login-code',
    ),
    url(
        r'^logout/',
        views.AuthTokenLogoutView.as_view(),
        name='authtoken-logout',
    )
]



from setuptools import find_packages, setup

from trench import __version__


setup(
    name='django-trench',
    version=__version__,
    packages=find_packages(exclude=['testproject', 'testproject.*']),
    include_package_data=True,
    license='MIT License',
    description='REST Multi-factor authentication package for Django',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/merixstudio/django-trench',
    author='Merixstudio',
    author_email='trench@merixstudio.com',
    install_requires=[
        'pyotp>=2.2.6',
        'twilio>=6.18.1',
        'yubico-client>=1.10.0',
        'smsapi-client>=2.2.5',
    ],
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _


class MFAMethod(models.Model):
    """
    Base model with MFA information linked to user.
    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        verbose_name=_('user'),
        related_name='mfa_methods',
    )
    name = models.CharField(
        _('name'),
        max_length=255,
    )
    secret = models.CharField(
        _('secret'),
        max_length=20,
    )
    is_primary = models.BooleanField(
        _('is primary'),
        default=False,
    )
    is_active = models.BooleanField(
        _('is active'),
        default=False,
    )
    _backup_codes = models.CharField(
        _('backup codes'),
        max_length=255,
        blank=True,
    )

    class Meta:
        verbose_name = _('MFA Method')
        verbose_name_plural = _('MFA Methods')

    def __str__(self):
        return '{} (User id: {})'.format(self.name, self.user_id)

    @property
    def backup_codes(self):
        return self._backup_codes.split(',')

    @backup_codes.setter
    def backup_codes(self, codes):
        self._backup_codes = ','.join(codes)

    def remove_backup_code(self, utilised_code):
        codes = self.backup_codes
        if utilised_code in codes:
            codes.remove(utilised_code)
            self.backup_codes = codes
            self.save()



from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework_jwt.utils import (
    jwt_encode_handler,
    jwt_payload_handler,
    jwt_response_payload_handler,
)

from trench.views.base import MFACodeLoginMixin, MFACredentialsLoginMixin


class ObtainJSONWebTokenMixin:
    def handle_user_login(self, request, serializer, *args, **kwargs):
        token = jwt_encode_handler(jwt_payload_handler(serializer.user))
        return Response(
            jwt_response_payload_handler(token, serializer.user, request)
        )


class JSONWebTokenLoginOrRequestMFACode(MFACredentialsLoginMixin,
                                        ObtainJSONWebTokenMixin,
                                        GenericAPIView):
    pass


class JSONWebTokenLoginWithMFACode(MFACodeLoginMixin,
                                   ObtainJSONWebTokenMixin,
                                   GenericAPIView):
    pass



from smtplib import SMTPException

from django.core.mail import send_mail
from django.utils.translation import ugettext_lazy as _

from trench.backends import AbstractMessageDispatcher
from trench.settings import api_settings


class SendMailBackend(AbstractMessageDispatcher):
    EMAIL_SUBJECT = _('Your verification code')

    def dispatch_message(self, *args, **kwargs):
        """
        Sends an email with verification code.
        :return:
        """

        code = self.create_code()

        try:
            send_mail(
                subject=self.EMAIL_SUBJECT,
                message=code,
                from_email=api_settings.FROM_EMAIL,
                recipient_list=[self.to],
                fail_silently=False,
            )
        except SMTPException:  # pragma: no cover
            return {
                'message': _('Email message with MFA code has not been sent.')
            }  # pragma: no cover

        return {'message': _('Email message with MFA code has been sent.')}



from django.conf.urls import url

from trench import views


__all__ = [
    'urlpatterns',
]

urlpatterns = [
    url(
        r'^(?P<method>[-\w]+)/activate/$',
        views.RequestMFAMethodActivationView.as_view(),
        name='mfa-activate',
    ),
    url(
        r'^(?P<method>[-\w]+)/activate/confirm/$',
        views.RequestMFAMethodActivationConfirmView.as_view(),
        name='mfa-activate-confirm',
    ),
    url(
        r'^(?P<method>[-\w]+)/deactivate/$',
        views.RequestMFAMethodDeactivationView.as_view(),
        name='mfa-deactivate',
    ),
    url(
        r'^(?P<method>[-\w]+)/codes/regenerate/$',
        views.RequestMFAMethodBackupCodesRegenerationView.as_view(),
        name='mfa-regenerate-codes',
    ),
    url(
        r'^code/request/$',
        views.RequestMFAMethodCode.as_view(),
        name='mfa-request-code',
    ),
    url(
        r'^mfa/config/$',
        views.GetMFAConfig.as_view(),
        name='mfa-config-info',
    ),
    url(
        r'^mfa/user-active-methods/$',
        views.ListUserActiveMFAMethods.as_view(),
        name='mfa-list-user-active-methods',
    ),
    url(
        r'^mfa/change-primary-method/$',
        views.ChangePrimaryMethod.as_view(),
        name='mfa-change-primary-method',
    ),
]
